import logging
from typing import Iterator, List, Optional, Tuple

from pydispix.autodraw import AutoDrawer
from pydispix.client import Client
from pydispix.color import Pixel

logger = logging.getLogger('pydispix')


class DistributedClient(Client):
    def __init__(
        self,
        token: Optional[str] = None,
        base_url: str = "https://pixels.pythondiscord.com/",
        *,
        total_tasks: int,
        controlled_tasks: List[int],
    ):
        """
        Add possibility to split tasks across multiple clients.

        `total_tasks` is the number of tasks our requests will be split to.

        `controlled_tasks` is the list of indices, defining which tasks are controlled here
        for example, with `total_tasks` = 10:
            First `MultiClient` (this machine): `controlled_tasks` = [0,1,2,3,4]
            Second `MultiClient` (collaborator): `controlled_tasks` = [5,6,7,8,9]

            Together, when running at the same time, and the tasks are split evenly.
            You usually want `total_tasks` to be equal to the number of your machines,
            since there is no real reason to give machines more than 1 controlled task,
            but as seen from the example, it is possible, if needed.
        """
        super().__init__(token, base_url)

        self.total_tasks = total_tasks
        self.controlled_tasks = controlled_tasks


class DistributedAutoDrawer(AutoDrawer):
    def __init__(self, client: DistributedClient, x: int, y: int, grid: List[List[Pixel]]):
        super().__init__(client, x, y, grid)
        # Redefine client for proper type highlights
        self.client: DistributedClient = client

    def _iter_coords(self) -> Iterator[Tuple[int, int]]:
        iter_coords = super()._iter_coords()
        width, _ = self.client.get_dimensions()
        for x, y in iter_coords:
            pixel_no = y * width + x
            task_no = pixel_no % self.client.total_tasks
            if task_no in self.client.controlled_tasks:
                yield x, y
            else:
                logger.debug(f"Skipping uncontrolled pixel ({x}, {y} - leaving for task {task_no}")



import re
from json.decoder import JSONDecodeError
from typing import Any, Optional, Union

import requests

from pydispix.ratelimits import RateLimitedEndpoint


class PyDisPixError(Exception):
    """Parent class for all exceptions defined by this library"""


class RateLimitBreached(PyDisPixError):
    """Request failed due to rate limit breach."""
    def __init__(self, *args, response: requests.Response, **kwargs):
        super().__init__(*args, **kwargs)

        # Get time limits from headers with RateLimitedEndpoint
        temp_rate_limit = RateLimitedEndpoint(response.url)
        temp_rate_limit.update_from_headers(response.headers)

        self.requests_limit = temp_rate_limit.requests_limit
        self.reset_time = temp_rate_limit.reset_time
        self.remaining_requests = temp_rate_limit.remaining_requests
        self.cooldown_time = temp_rate_limit.cooldown_time

        # Store the expected wait and the original response which trigerred this exception
        self.expected_wait_time = temp_rate_limit.get_wait_time()
        self.response = response

    def __str__(self):
        s = super().__str__()
        s += f"\nresponse={self.response.content}"
        if self.expected_wait_time != 0:
            s += f"\nexpected_wait_time={self.expected_wait_time}"

        return s


class InvalidToken(PyDisPixError, requests.HTTPError):
    """Invalid token used."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CanvasFormatError(PyDisPixError):
    """Exception raised when the canvas is badly formatted."""


class InvalidColor(PyDisPixError):
    """Invalid color format"""

    def __init__(self, *args, color: Any, **kwargs):
        super().__init__(*args, **kwargs)
        self.color = color

    def __str__(self) -> str:
        s = super().__str__()
        return s + f" color={self.color}"


class OutOfBoundaries(PyDisPixError):
    """Status code 422 - tried to draw a pixel outside of the canvas"""


def handle_invalid_body(response: requests.Response) -> Union[PyDisPixError, requests.HTTPError]:
    """
    Handle 442 (invalid body) error code. This code can mean many things,
    this function analyzes what exactly does the 442 refer to, and returns
    an appropriate exception for it.
    """
    if response.status_code != 422:
        raise ValueError("Invalid Body response must have 422 HTTP code.")

    detail = response.json()['detail']

    # Work with 1st entry only, we can't raise multiple errors anyway
    entry = detail[0]
    if entry["loc"][1] == "rgb":
        color = re.search(r"'(.+)' is not a valid color", entry["msg"]).groups()[0]
        return InvalidColor("Couldn't resolve color", color=color)
    if entry["loc"][1] in ("x", "y"):
        return OutOfBoundaries(entry["msg"])

    raise requests.HTTPError("Unrecognized 422 exception, please report this issue in the pydispix repository", response=response)


def get_response_result(
    exception: Union[requests.HTTPError, RateLimitBreached],
    key: Optional[str] = None,
    error_on_fail: bool = False
) -> Union[str, dict, list]:
    if not hasattr(exception, "response"):
        raise ValueError("This exception doesn't have a `response` attribute.")

    try:
        response = exception.response.json()
    except JSONDecodeError as exc:
        if error_on_fail:
            raise exc
        # Return the `content` message, this isn't JSON docodeable
        # If we can't decode to str text, don't bother with bytes, just raise `UnicodeDecodeError`
        return exception.response.content.decode("utf-8")

    if key is not None:
        try:
            response = response[key]
        except KeyError as exc:
            if error_on_fail:
                raise exc
            # Return the whole json, it doesn't contain wanted key
            return response

    return response



"""Tool for automatically drawing images."""
import logging
import time
from typing import Iterator, List, Optional, Tuple

import PIL.Image

from pydispix.canvas import Canvas, Pixel
from pydispix.client import Client
from pydispix.errors import OutOfBoundaries

logger = logging.getLogger('pydispix')


class AutoDrawer:
    """Tool for automatically drawing images."""

    def __init__(
        self,
        client: Client,
        x: int, y: int,
        grid: List[List[Pixel]]
    ):
        """Store the plan."""
        self.client = client
        self.grid = grid
        # Top left coords.
        self.x0 = x
        self.y0 = y
        # Bottom right coords.
        self.x1 = x + len(self.grid[0])
        self.y1 = y + len(self.grid)

        # Make sure we're within canvas boundaries
        canvas_width, canvas_height = canvas_size = self.client.get_dimensions()
        image_width, image_height = image_size = len(self.grid[0]), len(self.grid)
        if image_width > canvas_width or image_height > canvas_height:
            raise OutOfBoundaries(f"Can't draw picture bigger than the canvas ({image_size} > {canvas_size})")

    @staticmethod
    def _grid_from_img(
        image: PIL.Image.Image,
        scale: float = 1
    ):
        if image.mode == 'RGBA':
            new_image = PIL.Image.new('RGB', image.size)
            new_image.paste(image, mask=image)
            image = new_image

        width = round(image.width * scale)
        height = round(image.height * scale)

        if scale != 1:
            image = image.resize((width, height), PIL.Image.BILINEAR)

        data = list(image.getdata())
        grid = [
            [Pixel(r, g, b) for r, g, b in data[start:start + width]]
            for start in range(0, len(data), width)
        ]
        return grid

    @classmethod
    def load_image(
        cls,
        client: Client,
        xy: Tuple[int, int],
        image: PIL.Image.Image,
        scale: float = 1
    ) -> 'AutoDrawer':
        """Draw from the pixels of an image."""
        grid = cls._grid_from_img(image, scale)
        return cls(client, *xy, grid)

    def _iter_coords(self) -> Iterator[Tuple[int, int]]:
        """Iterate over the coordinates of the image."""
        for x in range(self.x0, self.x1):
            for y in range(self.y0, self.y1):
                yield x, y

    def draw_pixel(self, canvas: Canvas, x: int, y: int, show_progress: bool = True) -> bool:
        """
        Draw a pixel if not already drawn.

        Returns True if the pixel was not already drawn.
        """
        color = self.grid[y - self.y0][x - self.x0]
        if canvas[x, y] == color:
            logger.debug(f'Skipping already correct pixel at {x}, {y}.')
            return False
        self.client.put_pixel(x, y, color, show_progress=show_progress)
        return True

    def draw(self, guard: bool = False, guard_delay: int = 5, show_progress: bool = True):
        """Draw the pixels of the image, attempting each pixel max. once."""
        canvas = self.client.get_canvas()
        while True:
            for x, y in self._iter_coords():
                if self.draw_pixel(canvas, x, y, show_progress=show_progress):
                    canvas = self.client.get_canvas()

            if not guard:
                # Check this here, to act as do-while,
                # (always run first time, only continue if this is met)
                break
            # When we're guarding we need to update canvas even if no pixel was drawn
            # because otherwise we'd be looping over same non-updated canvas forever
            # since this looping with no changes takes a long time, we should also sleep
            # to avoid needless cpu usage
            time.sleep(guard_delay)
            canvas = self.client.get_canvas()


class MultiAutoDrawer:
    """Tool for automatically drawing set of images"""
    def __init__(
        self,
        client: Client,
        positions: List[Tuple[int, int]],
        grids: List[List[List[Pixel]]],
        one_by_one: bool = True
    ):
        self.client = client
        self.drawers = [
            AutoDrawer(client, *position, grid)
            for position, grid in zip(positions, grids)
        ]
        self.positions_generator = self._one_by_one_positions if one_by_one else self._per_pixel_positions

    @classmethod
    def load_images(
        cls,
        client: Client,
        positions: List[Tuple[int, int]],
        images: List[PIL.Image.Image],
        scales: Optional[List[int]] = None,
        one_by_one: bool = True
    ) -> "MultiAutoDrawer":
        """Draw from pixels on the images."""
        if scales is None:
            # Default all scales to 1
            scales = [1 for _ in images]

        grids = [
            AutoDrawer._grid_from_img(image, scale)
            for image, scale in zip(images, scales)
        ]

        return cls(client, positions, grids, one_by_one)

    def _one_by_one_positions(self) -> Iterator[Tuple[AutoDrawer, Tuple[int, int]]]:
        """
        Return iterator of the pixels to check for the passed set of grids (images).
        This will one by one through the individual images. This allows for prioritizing
        certain images over others.
        """
        coord_generators = [drawer._iter_coords() for drawer in self.drawers]
        for drawer, coord_generator in zip(self.drawers, coord_generators):
            for x, y in coord_generator:
                yield drawer, (x, y)

    def _per_pixel_positions(self) -> Iterator[Tuple[AutoDrawer, Tuple[int, int]]]:
        """
        This is similar to `_one_by_one_positions`, except instead of iterating one by one,
        we are instead iterating through individual pixels, i.e. all 1st pixels from
        all images, all 2nd pixels, etc.
        """
        coord_generators = [drawer._iter_coords() for drawer in self.drawers]

        # Keep track of which drawers are already depleted
        depleted = {drawer: False for drawer in self.drawers}
        # Keep running as long as we have any non-depleted drawers
        while any(is_depleted is False for is_depleted in depleted.values()):
            for drawer, coord_generator in zip(self.drawers, coord_generators):
                try:
                    x, y = next(coord_generator)
                except StopIteration:
                    # If we already depleted pixels from this drawer, just ignore
                    # it and move on to another one
                    depleted[drawer] = True
                    continue
                yield drawer, (x, y)

    def draw(
        self,
        guard: bool = False,
        guard_delay: int = 5,
        show_progress: bool = True,
    ):
        canvas = self.client.get_canvas()

        while True:
            for drawer, (x, y) in self.positions_generator():
                if drawer.draw_pixel(canvas, x, y, show_progress=show_progress):
                    canvas = self.client.get_canvas()
            if not guard:
                # Check this here, to act as do-while,
                # (always run first time, only continue if this is met)
                break
            # When we're guarding we need to update canvas even if no pixel was drawn
            # because otherwise we'd be looping over same non-updated canvas forever
            # since this looping with no changes takes a long time, we should also sleep
            # to avoid needless cpu usage
            time.sleep(guard_delay)
            canvas = self.client.get_canvas()
